@@ -1,11 +1,13 @@
 #include <ctype.h>
 #include <string.h>
+#include "backports.h"
 #include "checks.h"
-#include "is_integerish.h"
+#include "integerish.h"
+#include "is_sorted.h"
 #include "any_missing.h"
 #include "any_infinite.h"
 #include "all_missing.h"
-#include "all_nchar.h"
+#include "find_min_nchar.h"
 #include "helper.h"
 #include "guess_type.h"
 
@@ -30,6 +32,23 @@ static char msg[255] = "";
         } \
     }
 
+#define HANDLE_INTEGERISH_NULL(tol, null_ok) \
+    if (isNull((x))) { \
+        if (asFlag((null_ok), "null.ok")) \
+            return ScalarLogical(TRUE); \
+        snprintf(msg, 255, "Must be of type 'integerish', not 'NULL'"); \
+        return ScalarString(mkChar(msg)); \
+    } else { \
+        int_err_t ok = checkIntegerish(x, dtol, FALSE); \
+        switch(ok.err) { \
+            case INT_OK: break; \
+            case INT_TYPE: snprintf(msg, 255, "Must be of type 'integerish'%s, not '%s'", asFlag(null_ok, "null_ok") ? " (or 'NULL')" : "", guess_type(x)); return ScalarString(mkChar(msg)); \
+            case INT_RANGE: snprintf(msg, 255, "Must be of type 'integerish', but element %ld is not in integer range", ok.pos); return ScalarString(mkChar(msg)); \
+            case INT_TOL: snprintf(msg, 255, "Must be of type 'integerish', but element %ld is not close to an integer", ok.pos); return ScalarString(mkChar(msg)); \
+            case INT_COMPLEX: snprintf(msg, 255, "Must be of type 'integerish', but element %ld has an imaginary part", ok.pos); return ScalarString(mkChar(msg)); \
+        } \
+    }
+
 #define HANDLE_NA(x, na_ok) \
     if (is_scalar_na((x))) { \
         if (asFlag((na_ok), "na.ok")) \
@@ -43,7 +62,6 @@ static char msg[255] = "";
     UNPROTECT((p)); \
     if (!TMP) return ScalarString(mkChar(msg));
 
-
 /*********************************************************************************************************************/
 /* Some helpers                                                                                                      */
 /*********************************************************************************************************************/
@@ -63,48 +81,117 @@ static SEXP result(const char *fmt, ...) {
     return ScalarString(mkChar(msg));
 }
 
+static void fmt_posixct(char * out, SEXP x) {
+    SEXP call = PROTECT(allocVector(LANGSXP, 2));
+    SETCAR(call, install("format.POSIXct"));
+    SETCADR(call, x);
+    SEXP result = PROTECT(eval(call, R_GlobalEnv));
+
+    strncpy(out, CHAR(STRING_ELT(result, 0)), 255);
+    out[255] = '\0';
+    UNPROTECT(2);
+}
+
 static Rboolean check_bounds(SEXP x, SEXP lower, SEXP upper) {
     double tmp = asNumber(lower, "lower");
     if (R_FINITE(tmp)) {
+        const R_xlen_t n = length(x);
         if (isReal(x)) {
-            const double *xp = REAL(x);
-            const double * const xend = xp + xlength(x);
-            for (; xp != xend; xp++) {
-                if (!ISNAN(*xp) && *xp < tmp)
-                    return message("All elements must be >= %g", tmp);
+            const double *xp = REAL_RO(x);
+            for (R_xlen_t i = 0; i < n; i++) {
+                if (!ISNAN(xp[i]) && xp[i] < tmp)
+                    return message("Element %i is not >= %g", i + i, tmp);
             }
         } else if (isInteger(x)) {
-            const int *xp = INTEGER(x);
-            const int * const xend = xp + xlength(x);
-            for (; xp != xend; xp++) {
-                if (*xp != NA_INTEGER && *xp < tmp)
-                    return message("All elements must be >= %g", tmp);
+            const int *xp = INTEGER_RO(x);
+            for (R_xlen_t i = 0; i < n; i++) {
+                if (xp[i] != NA_INTEGER && xp[i] < tmp)
+                    return message("Element %i is not >= %g", i + 1, tmp);
             }
         }
     }
 
     tmp = asNumber(upper, "upper");
     if (R_FINITE(tmp)) {
+        const R_xlen_t n = length(x);
         if (isReal(x)) {
-            const double *xp = REAL(x);
-            const double * const xend = xp + xlength(x);
-            for (; xp != xend; xp++) {
-                if (!ISNAN(*xp) && *xp > tmp)
-                    return message("All elements must be <= %g", tmp);
+            const double *xp = REAL_RO(x);
+            for (R_xlen_t i = 0; i < n; i++) {
+                if (!ISNAN(xp[i]) && xp[i] > tmp)
+                    return message("Element %i is not <= %g", i + 1, tmp);
             }
         } else if (isInteger(x)) {
-            const int *xp = INTEGER(x);
-            const int * const xend = xp + xlength(x);
-            for (; xp != xend; xp++) {
-                if (*xp != NA_INTEGER && *xp > tmp)
-                    return message("All elements must be <= %g", tmp);
+            const int *xp = INTEGER_RO(x);
+            for (R_xlen_t i = 0; i < n; i++) {
+                if (xp[i] != NA_INTEGER && xp[i] > tmp)
+                    return message("Element %i is not <= %g", i + 1, tmp);
             }
         }
     }
     return TRUE;
 }
 
-static Rboolean check_strict_names(SEXP x) {
+
+static Rboolean check_posix_bounds(SEXP x, SEXP lower, SEXP upper) {
+    if (isNull(lower) && isNull(upper))
+        return TRUE;
+
+    SEXP tz = PROTECT(getAttrib(x, install("tzone")));
+    const Rboolean null_tz = isNull(tz);
+
+    if (!isNull(lower)) {
+        if (!is_class_posixct(lower) || length(lower) != 1)
+            error("Argument 'lower' must be provided as single POSIXct time");
+        SEXP lower_tz = PROTECT(getAttrib(lower, install("tzone")));
+        if (null_tz != isNull(lower_tz) ||
+            (!null_tz && !isNull(lower_tz) && strcmp(CHAR(STRING_ELT(tz, 0)), CHAR(STRING_ELT(lower_tz, 0))) != 0)) {
+            UNPROTECT(2);
+            return message("Timezones of 'x' and 'lower' must match");
+        }
+
+        const double tmp = REAL_RO(lower)[0];
+        const double *xp = REAL_RO(x);
+        const R_xlen_t n = length(x);
+        for (R_xlen_t i = 0; i < n; i++) {
+            if (!ISNAN(xp[i]) && xp[i] < tmp) {
+                char fmt[256];
+                fmt_posixct(fmt, lower);
+                UNPROTECT(2);
+                return message("Element %i is not >= %s", i + 1, fmt);
+            }
+        }
+        UNPROTECT(1);
+    }
+
+    if (!isNull(upper)) {
+        if (!is_class_posixct(upper) || length(upper) != 1)
+            error("Argument 'upper' must be provided as single POSIXct time");
+        SEXP upper_tz = PROTECT(getAttrib(upper, install("tzone")));
+        if (null_tz != isNull(upper_tz) ||
+            (!null_tz && !isNull(upper_tz) && strcmp(CHAR(STRING_ELT(tz, 0)), CHAR(STRING_ELT(upper_tz, 0))) != 0)) {
+            UNPROTECT(2);
+            return message("Timezones of 'x' and 'upper' must match");
+        }
+
+        const double tmp = REAL_RO(upper)[0];
+        const double *xp = REAL_RO(x);
+        const R_xlen_t n = length(x);
+        for (R_xlen_t i = 0; i < n; i++) {
+            if (!ISNAN(xp[i]) && xp[i] > tmp) {
+                char fmt[256];
+                fmt_posixct(fmt, upper);
+                UNPROTECT(2);
+                return message("Element %i is not <= %s", i + 1, fmt);
+            }
+        }
+        UNPROTECT(1);
+    }
+
+    UNPROTECT(1);
+    return TRUE;
+}
+
+static R_xlen_t check_strict_names(SEXP x) {
     const R_xlen_t nx = xlength(x);
     const char *str;
     for (R_xlen_t i = 0; i < nx; i++) {
@@ -112,13 +199,13 @@ static Rboolean check_strict_names(SEXP x) {
         while (*str == '.')
             str++;
         if (!isalpha(*str))
-            return FALSE;
+            return i + 1;
         for (; *str != '\0'; str++) {
             if (!isalnum(*str) && *str != '.' && *str != '_')
-                return FALSE;
+                return i + 1;
         }
     }
-    return TRUE;
+    return 0;
 }
 
 static Rboolean check_names(SEXP nn, const char * type, const char * what) {
@@ -138,13 +225,28 @@ static Rboolean check_names(SEXP nn, const char * type, const char * what) {
         error("Unknown type '%s' to specify check for names. Supported are 'unnamed', 'named', 'unique' and 'strict'.", type);
     }
 
-    if (isNull(nn) || any_missing_string(nn) || !all_nchar(nn, 1, FALSE))
-        return message("%s must be named", what);
+    if (isNull(nn)) {
+        return message("%s must be named, but is NULL", what);
+    }
+
+    R_xlen_t pos = find_missing_string(nn);
+    if (pos > 0) {
+        return message("%s must be named, but name %i is NA", what, pos);
+    }
+
+    pos = find_min_nchar(nn, 1, FALSE);
+    if (pos > 0) {
+        return message("%s must be named, but name %i is empty", what, pos);
+    }
+
     if (checks >= T_UNIQUE) {
-        if (any_duplicated(nn, FALSE) != 0)
-            return message("%s must be uniquely named", what);
-        if (checks >= T_STRICT && !check_strict_names(nn)) {
-            return message("%s must be named according to R's variable naming conventions and may not contain special characters", what);
+        pos = any_duplicated(nn, FALSE);
+        if (pos > 0)
+            return message("%s must be uniquely named, but name %i is duplicated", what, pos);
+        if (checks >= T_STRICT) {
+            pos = check_strict_names(nn);
+            if (pos > 0)
+                return message("%s must be named according to R's variable naming conventions and may not contain special characters", what);
         }
     }
     return TRUE;
@@ -177,16 +279,22 @@ static Rboolean check_vector_len(SEXP x, SEXP len, SEXP min_len, SEXP max_len) {
 }
 
 static Rboolean check_vector_missings(SEXP x, SEXP any_missing, SEXP all_missing) {
-    if (!asFlag(any_missing, "any.missing") && any_missing_atomic(x))
-        return message("Contains missing values");
+    if (!asFlag(any_missing, "any.missing")) {
+        R_xlen_t pos = find_missing_vector(x);
+        if (pos > 0)
+            return message("Contains missing values (element %i)", pos);
+    }
     if (!asFlag(all_missing, "all.missing") && all_missing_atomic(x))
         return message("Contains only missing values");
     return TRUE;
 }
 
 static Rboolean check_vector_unique(SEXP x, SEXP unique) {
-    if (asFlag(unique, "unique") && any_duplicated(x, FALSE) > 0)
-        return message("Contains duplicated values");
+    if (asFlag(unique, "unique")) {
+        R_xlen_t pos = any_duplicated(x, FALSE);
+        if (pos > 0)
+            return message("Contains duplicated values, position %i", pos);
+    }
     return TRUE;
 }
 
@@ -197,32 +305,43 @@ static Rboolean check_vector_names(SEXP x, SEXP names) {
 }
 
 static Rboolean check_vector_finite(SEXP x, SEXP finite) {
+    // FIXME: pos
     if (asFlag(finite, "finite") && any_infinite(x))
         return message("Must be finite");
     return TRUE;
 }
 
-static Rboolean check_matrix_dims(SEXP x, SEXP min_rows, SEXP min_cols, SEXP rows, SEXP cols) {
-    if (!isNull(min_rows) || !isNull(rows)) {
+static Rboolean check_matrix_dims(SEXP x, SEXP min_rows, SEXP max_rows, SEXP min_cols, SEXP max_cols, SEXP rows, SEXP cols) {
+    if (!isNull(min_rows) || !isNull(max_rows) || !isNull(rows)) {
         R_len_t xrows = get_nrows(x);
         if (!isNull(min_rows)) {
             R_len_t cmp = asLength(min_rows, "min.rows");
             if (xrows < cmp)
                 return message("Must have at least %i rows, but has %i rows", cmp, xrows);
         }
+        if (!isNull(max_rows)) {
+            R_len_t cmp = asLength(max_rows, "max.rows");
+            if (xrows > cmp)
+                return message("Must have at most %i rows, but has %i rows", cmp, xrows);
+        }
         if (!isNull(rows)) {
             R_len_t cmp = asLength(rows, "rows");
             if (xrows != cmp)
                 return message("Must have exactly %i rows, but has %i rows", cmp, xrows);
         }
     }
-    if (!isNull(min_cols) || !isNull(cols)) {
+    if (!isNull(min_cols) || !isNull(max_cols) || !isNull(cols)) {
         R_len_t xcols = get_ncols(x);
         if (!isNull(min_cols)) {
             R_len_t cmp = asLength(min_cols, "min.cols");
             if (xcols < cmp)
                 return message("Must have at least %i cols, but has %i cols", cmp, xcols);
         }
+        if (!isNull(max_cols)) {
+            R_len_t cmp = asLength(max_cols, "max.cols");
+            if (xcols > cmp)
+                return message("Must have at most %i cols, but has %i cols", cmp, xcols);
+        }
         if (!isNull(cols)) {
             R_len_t cmp = asCount(cols, "cols");
             if (xcols != cmp)
@@ -248,7 +367,7 @@ static Rboolean check_storage(SEXP x, SEXP mode) {
             if (!isIntegerish(x, INTEGERISH_DEFAULT_TOL, FALSE))
                 return message("Must store integerish values");
         } else if (strcmp(storage, "numeric") == 0) {
-            if (!isStrictlyNumeric(x))
+            if (!is_class_numeric(x))
                 return message("Must store numerics");
         } else if (strcmp(storage, "complex") == 0) {
             if (!isComplex(x))
@@ -257,7 +376,7 @@ static Rboolean check_storage(SEXP x, SEXP mode) {
             if (!isString(x))
                 return message("Must store characters");
         } else if (strcmp(storage, "list") == 0) {
-            if (!isRList(x))
+            if (!is_class_list(x))
                 return message("Must store a list");
         } else if (strcmp(storage, "atomic") == 0) {
             if (!isVectorAtomic(x))
@@ -272,67 +391,18 @@ static Rboolean check_storage(SEXP x, SEXP mode) {
 static inline Rboolean is_scalar_na(SEXP x) {
     if (xlength(x) == 1) {
         switch(TYPEOF(x)) {
-            case LGLSXP: return (LOGICAL(x)[0] == NA_LOGICAL);
-            case INTSXP: return (INTEGER(x)[0] == NA_INTEGER);
-            case REALSXP: return ISNAN(REAL(x)[0]);
+            case LGLSXP: return (LOGICAL_RO(x)[0] == NA_LOGICAL);
+            case INTSXP: return (INTEGER_RO(x)[0] == NA_INTEGER);
+            case REALSXP: return ISNAN(REAL_RO(x)[0]);
             case STRSXP: return (STRING_ELT(x, 0) == NA_STRING);
         }
     }
     return FALSE;
 }
 
-
-static Rboolean is_sorted_integer(SEXP x) {
-    R_xlen_t i = 0;
-    const int * const xi = INTEGER(x);
-    const R_xlen_t n = xlength(x);
-    while(xi[i] == NA_INTEGER) {
-        i++;
-        if (i == n)
-            return TRUE;
-    }
-
-    for (R_xlen_t j = i + 1; j < n; j++) {
-        if (xi[j] != NA_INTEGER) {
-            if (xi[i] > xi[j])
-                return FALSE;
-            i = j;
-        }
-    }
-    return TRUE;
-}
-
-
-static Rboolean is_sorted_double(SEXP x) {
-    R_xlen_t i = 0;
-    const double * const xr = REAL(x);
-    const R_xlen_t n = xlength(x);
-    while(xr[i] == NA_REAL) {
-        i++;
-        if (i == n)
-            return TRUE;
-    }
-
-    for (R_xlen_t j = i + 1; j < n; j++) {
-        if (xr[j] != NA_REAL) {
-            if (xr[i] > xr[j])
-                return FALSE;
-            i = j;
-        }
-    }
-    return TRUE;
-}
-
-
 static Rboolean check_vector_sorted(SEXP x, SEXP sorted) {
     if (asFlag(sorted, "sorted") && xlength(x) > 1) {
-        Rboolean ok;
-        switch(TYPEOF(x)) {
-            case INTSXP: ok = is_sorted_integer(x); break;
-            case REALSXP: ok = is_sorted_double(x); break;
-            default: error("Checking for sorted vector only possible for integer and double");
-        }
-        if (!ok)
+        if (!isSorted(x))
             return message("Must be sorted");
     }
     return TRUE;
@@ -341,22 +411,23 @@ static Rboolean check_vector_sorted(SEXP x, SEXP sorted) {
 /*********************************************************************************************************************/
 /* Exported check functions                                                                                          */
 /*********************************************************************************************************************/
-SEXP attribute_hidden c_check_character(SEXP x, SEXP min_chars, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isString(x) || all_missing_atomic(x), "character", null_ok);
+SEXP attribute_hidden c_check_character(SEXP x, SEXP min_chars, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP sorted, SEXP names, SEXP null_ok) {
+    HANDLE_TYPE_NULL(is_class_string(x) || all_missing_atomic(x), "character", null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
     if (!isNull(min_chars)) {
         R_xlen_t n = asCount(min_chars, "min.chars");
-        if (n > 0 && !all_nchar(x, n, TRUE))
+        if (n > 0 && find_min_nchar(x, n, TRUE) > 0)
             return result("All elements must have at least %i characters", n);
     }
     ASSERT_TRUE(check_vector_unique(x, unique));
+    ASSERT_TRUE(check_vector_sorted(x, sorted));
     return ScalarLogical(TRUE);
 }
 
 SEXP attribute_hidden c_check_complex(SEXP x, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isComplex(x) || all_missing_atomic(x), "complex", null_ok);
+    HANDLE_TYPE_NULL(is_class_complex(x) || all_missing_atomic(x), "complex", null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -364,28 +435,39 @@ SEXP attribute_hidden c_check_complex(SEXP x, SEXP any_missing, SEXP all_missing
     return ScalarLogical(TRUE);
 }
 
-SEXP attribute_hidden c_check_dataframe(SEXP x, SEXP any_missing, SEXP all_missing, SEXP min_rows, SEXP min_cols, SEXP rows, SEXP cols, SEXP row_names, SEXP col_names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isFrame(x), "data.frame", null_ok);
-    ASSERT_TRUE(check_matrix_dims(x, min_rows, min_cols, rows, cols));
+SEXP attribute_hidden c_check_dataframe(SEXP x, SEXP any_missing, SEXP all_missing, SEXP min_rows, SEXP max_rows, SEXP min_cols, SEXP max_cols, SEXP rows, SEXP cols, SEXP row_names, SEXP col_names, SEXP null_ok) {
+    HANDLE_TYPE_NULL(is_class_frame(x), "data.frame", null_ok);
+    ASSERT_TRUE(check_matrix_dims(x, min_rows, max_rows, min_cols, max_cols, rows, cols));
 
     if (!isNull(row_names)) {
         SEXP nn = PROTECT(getAttrib(x, install("row.names")));
-        if (isInteger(nn))
-            nn = coerceVector(nn, STRSXP);
-        ASSERT_TRUE_UNPROTECT(check_names(nn, asString(row_names, "row.names"), "Rows"), 1);
+        int nprotect = 1;
+        if (isInteger(nn)) {
+            nn = PROTECT(coerceVector(nn, STRSXP));
+            nprotect++;
+        }
+        ASSERT_TRUE_UNPROTECT(check_names(nn, asString(row_names, "row.names"), "Rows"), nprotect);
     }
 
-    if (!isNull(col_names))
+    if (!isNull(col_names)) {
         ASSERT_TRUE(check_named(x, asString(col_names, "col.names"), "Columns"));
-    if (!asFlag(any_missing, "any.missing") && any_missing_frame(x))
-        return result("Contains missing values");
-    if (!asFlag(all_missing, "all.missing") && all_missing_frame(x))
+    }
+    if (!asFlag(any_missing, "any.missing")) {
+        pos2d_t pos = find_missing_frame(x);
+        if (pos.i > 0) {
+            const char * nn = CHAR(STRING_ELT(getAttrib(x, R_NamesSymbol), pos.j));
+            return result("Contains missing values (column '%s', row %i)", nn, pos.i);
+        }
+    }
+
+    if (!asFlag(all_missing, "all.missing") && all_missing_frame(x)) {
         return result("Contains only missing values");
+    }
     return ScalarLogical(TRUE);
 }
 
 SEXP attribute_hidden c_check_factor(SEXP x, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isFactor(x) || all_missing_atomic(x), "factor", null_ok);
+    HANDLE_TYPE_NULL(is_class_factor(x) || all_missing_atomic(x), "factor", null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -394,7 +476,7 @@ SEXP attribute_hidden c_check_factor(SEXP x, SEXP any_missing, SEXP all_missing,
 }
 
 SEXP attribute_hidden c_check_integer(SEXP x, SEXP lower, SEXP upper, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP sorted, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isInteger(x) || all_missing_atomic(x), "integer", null_ok);
+    HANDLE_TYPE_NULL(is_class_integer(x) || all_missing_atomic(x), "integer", null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -406,7 +488,7 @@ SEXP attribute_hidden c_check_integer(SEXP x, SEXP lower, SEXP upper, SEXP any_m
 
 SEXP attribute_hidden c_check_integerish(SEXP x, SEXP tol, SEXP lower, SEXP upper, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP sorted, SEXP names, SEXP null_ok) {
     double dtol = asNumber(tol, "tol");
-    HANDLE_TYPE_NULL(isIntegerish(x, dtol, FALSE) || all_missing_atomic(x), "integerish", null_ok);
+    HANDLE_INTEGERISH_NULL(dtol, null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -417,7 +499,7 @@ SEXP attribute_hidden c_check_integerish(SEXP x, SEXP tol, SEXP lower, SEXP uppe
 }
 
 SEXP attribute_hidden c_check_list(SEXP x, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isRList(x), "list", null_ok)
+    HANDLE_TYPE_NULL(is_class_list(x), "list", null_ok)
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -426,7 +508,7 @@ SEXP attribute_hidden c_check_list(SEXP x, SEXP any_missing, SEXP all_missing, S
 }
 
 SEXP attribute_hidden c_check_logical(SEXP x, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isLogical(x) || all_missing_atomic(x), "logical", null_ok);
+    HANDLE_TYPE_NULL(is_class_logical(x) || all_missing_atomic(x), "logical", null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -434,10 +516,10 @@ SEXP attribute_hidden c_check_logical(SEXP x, SEXP any_missing, SEXP all_missing
     return ScalarLogical(TRUE);
 }
 
-SEXP attribute_hidden c_check_matrix(SEXP x, SEXP mode, SEXP any_missing, SEXP all_missing, SEXP min_rows, SEXP min_cols, SEXP rows, SEXP cols, SEXP row_names, SEXP col_names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isMatrix(x), "matrix", null_ok);
+SEXP attribute_hidden c_check_matrix(SEXP x, SEXP mode, SEXP any_missing, SEXP all_missing, SEXP min_rows, SEXP max_rows, SEXP min_cols, SEXP max_cols, SEXP rows, SEXP cols, SEXP row_names, SEXP col_names, SEXP null_ok) {
+    HANDLE_TYPE_NULL(is_class_matrix(x), "matrix", null_ok);
     ASSERT_TRUE(check_storage(x, mode));
-    ASSERT_TRUE(check_matrix_dims(x, min_rows, min_cols, rows, cols));
+    ASSERT_TRUE(check_matrix_dims(x, min_rows, max_rows, min_cols, max_cols, rows, cols));
 
     if (!isNull(row_names) && xlength(x) > 0) {
         SEXP nn = PROTECT(getAttrib(x, R_DimNamesSymbol));
@@ -452,15 +534,25 @@ SEXP attribute_hidden c_check_matrix(SEXP x, SEXP mode, SEXP any_missing, SEXP a
             nn = VECTOR_ELT(nn, 1);
         ASSERT_TRUE_UNPROTECT(check_names(nn, asString(col_names, "col.names"), "Columns"), 1);
     }
-    ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
+
+    if (!asFlag(any_missing, "any.missing")) {
+        pos2d_t pos = find_missing_matrix(x);
+        if (pos.i > 0) {
+            return result("Contains missing values (row %i, col %i)", pos.i, pos.j);
+        }
+    }
+
+    if (!asFlag(all_missing, "all.missing") && all_missing_atomic(x)) {
+        return result("Contains only missing values");
+    }
     return ScalarLogical(TRUE);
 }
 
 SEXP attribute_hidden c_check_array(SEXP x, SEXP mode, SEXP any_missing, SEXP d, SEXP min_d, SEXP max_d, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isArray(x), "array", null_ok);
+    HANDLE_TYPE_NULL(is_class_array(x), "array", null_ok);
     ASSERT_TRUE(check_storage(x, mode));
 
-    if (!asFlag(any_missing, "any.missing") && any_missing_atomic(x))
+    if (!asFlag(any_missing, "any.missing") && find_missing_vector(x) > 0)
         return result("Contains missing values");
 
     R_len_t ndim = length(getAttrib(x, R_DimSymbol));
@@ -500,7 +592,19 @@ SEXP attribute_hidden c_check_names(SEXP x, SEXP type) {
 
 
 SEXP attribute_hidden c_check_numeric(SEXP x, SEXP lower, SEXP upper, SEXP finite, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP sorted, SEXP names, SEXP null_ok) {
-    HANDLE_TYPE_NULL(isStrictlyNumeric(x) || all_missing_atomic(x), "numeric", null_ok);
+    HANDLE_TYPE_NULL(is_class_numeric(x) || all_missing_atomic(x), "numeric", null_ok);
+    ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
+    ASSERT_TRUE(check_vector_names(x, names));
+    ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
+    ASSERT_TRUE(check_bounds(x, lower, upper));
+    ASSERT_TRUE(check_vector_finite(x, finite));
+    ASSERT_TRUE(check_vector_unique(x, unique));
+    ASSERT_TRUE(check_vector_sorted(x, sorted));
+    return ScalarLogical(TRUE);
+}
+
+SEXP attribute_hidden c_check_double(SEXP x, SEXP lower, SEXP upper, SEXP finite, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP sorted, SEXP names, SEXP null_ok) {
+    HANDLE_TYPE_NULL(is_class_double(x) || all_missing_atomic(x), "double", null_ok);
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -524,8 +628,15 @@ SEXP attribute_hidden c_check_vector(SEXP x, SEXP strict, SEXP any_missing, SEXP
     return ScalarLogical(TRUE);
 }
 
+SEXP attribute_hidden c_check_raw(SEXP x, SEXP len, SEXP min_len, SEXP max_len, SEXP names, SEXP null_ok) {
+    HANDLE_TYPE_NULL(is_class_raw(x), "raw", null_ok);
+    ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
+    ASSERT_TRUE(check_vector_names(x, names));
+    return ScalarLogical(TRUE);
+}
+
 SEXP attribute_hidden c_check_atomic(SEXP x, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names) {
-    HANDLE_TYPE(isNull(x) || isVectorAtomic(x), "atomic");
+    HANDLE_TYPE(is_class_atomic(x), "atomic");
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -534,7 +645,7 @@ SEXP attribute_hidden c_check_atomic(SEXP x, SEXP any_missing, SEXP all_missing,
 }
 
 SEXP attribute_hidden c_check_atomic_vector(SEXP x, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP names) {
-    HANDLE_TYPE(isAtomicVector(x), "atomic vector");
+    HANDLE_TYPE(is_class_atomic_vector(x), "atomic vector");
     ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
     ASSERT_TRUE(check_vector_names(x, names));
     ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
@@ -574,7 +685,7 @@ SEXP attribute_hidden c_check_int(SEXP x, SEXP na_ok, SEXP lower, SEXP upper, SE
 
 SEXP attribute_hidden c_check_number(SEXP x, SEXP na_ok, SEXP lower, SEXP upper, SEXP finite, SEXP null_ok) {
     HANDLE_NA(x, na_ok);
-    HANDLE_TYPE_NULL(isStrictlyNumeric(x), "number", null_ok);
+    HANDLE_TYPE_NULL(is_class_numeric(x), "number", null_ok);
     if (xlength(x) != 1)
         return result("Must have length 1");
     ASSERT_TRUE(check_vector_finite(x, finite));
@@ -589,7 +700,7 @@ SEXP attribute_hidden c_check_string(SEXP x, SEXP na_ok, SEXP min_chars, SEXP nu
         return result("Must have length 1");
     if (!isNull(min_chars)) {
         R_xlen_t n = asCount(min_chars, "min.chars");
-        if (!all_nchar(x, n, TRUE))
+        if (find_min_nchar(x, n, TRUE) > 0)
             return result("Must have at least %i characters", n);
     }
 
@@ -604,6 +715,17 @@ SEXP attribute_hidden c_check_scalar(SEXP x, SEXP na_ok, SEXP null_ok) {
     return ScalarLogical(TRUE);
 }
 
+SEXP attribute_hidden c_check_posixct(SEXP x, SEXP lower, SEXP upper, SEXP any_missing, SEXP all_missing, SEXP len, SEXP min_len, SEXP max_len, SEXP unique, SEXP sorted, SEXP null_ok) {
+    HANDLE_TYPE_NULL(is_class_posixct(x), "POSIXct", null_ok);
+    ASSERT_TRUE(check_vector_len(x, len, min_len, max_len));
+    ASSERT_TRUE(check_vector_missings(x, any_missing, all_missing));
+    ASSERT_TRUE(check_vector_unique(x, unique));
+    ASSERT_TRUE(check_posix_bounds(x, lower, upper));
+    ASSERT_TRUE(check_vector_sorted(x, sorted));
+
+    return ScalarLogical(TRUE);
+}
+
 #undef HANDLE_TYPE
 #undef HANDLE_TYPE_NULL
 #undef HANDLE_NA
